/*
 * USBLooper    v0.9.0    1999-07-17    Thomas S. Winzig
 * 
 * -----------------------------------------------------------------------
 *
 * == OVERVIEW ==
 *
 * This application hides in the background waiting for USB devices to 
 * connect/disconnect. If you want these notices as well, just send it a 
 * message saying so:
 *
 * BMessenger msgr(USB_LOOPER_SIG);
 * msgr.SendMessage(USB_LOOPER_HELLO);
 *
 * (Don't forget to #include "USBLooperUtil.h".)
 *
 * You will now receive messages whenever a USB device is connected 
 * (USB_DEVICE_CONNECTED) or disconnected (USB_DEVICE_DISCONNECTED).
 * You might also receive a message if USBLooper has an unrecoverable
 * error, and can no longer be reliably used (USB_LOOPER_KABOOM).
 *
 * The device connect/disconnect messages will have 'name' and 'location'
 * string fields, where 'name' is the USB device name (as given by the
 * device), and 'location' is the full path to the device. E.g.:
 * "/dev/bus/usb/0/1/1". You can then decide whether the device interests
 * you, and if so, use the USBKit (from Be) to communicate with it. (The
 * USBDevice class in particular.)
 *
 * If you want to know what USB devices are already connected, send
 * USBLooper a message (USB_DEVICES_ALREADY_CONNECTED) asking for that 
 * info, and it will send you a reply (USB_DEVICES). This reply has
 * multiple strings stored (like a hashtable), where the name is the
 * name of the device, and the value is the full path to the device. Here 
 * is an example that looks to see if a Kodak DC265 camera is connected:
 *
 *	BMessage reply;
 *	msgr.SendMessage(new BMessage(USB_DEVICES_ALREADY_CONNECTED), &reply);
 *	if (reply.what == USB_DEVICES && reply.HasString("KODAK DC265 ZOOM DIGITAL CA"))
 *	{
 *		cout << "Kodak DC265 Zoom digital camera is connected at ";
 *		cout << reply.FindString("KODAK DC265 ZOOM DIGITAL CA") << endl;
 *		// It's possible the same type of device is connected more than once,
 *		// so you should use reply.GetInfo(...) to find out which one you want.
 *	}
 *
 * -----------------------------------------------------------------------
 *
 * You will probably want to install USBLooper into /boot/home/config/bin,
 * and then add it to your UserBootscript so that it starts up immediately. 
 *
 * -----------------------------------------------------------------------
 *
 * == LICENSE / ETC ==
 *
 * In the interest of encouraging a standard way for USB application
 * writers to access USB devices, I am releasing this application and
 * source code under the LGPL (Lesser General Public License). In essence, 
 * this allows you to use/modify this library in a commercial setting, 
 * with only a few restrictions. Please read the LGPL to which you are bound:
 *
 * http://www.gnu.org/copyleft/lesser.html
 *
 * (Notice the application signature for USBLooper has 'LGPL' as the vendor.)
 *
 * USBLooper would be useless without the cool C++ USB interface created 
 * by Brian Swetland (swetland@be.com). Thanks Brian!
 *
 * If you like USBLooper, feel free to show your appreciation by squashing
 * bugs, offering improvements, and of course, writing USB apps for BeOS! I'd
 * especially love to see a driver for my Microsoft digital USB speakers! :-)
 * For information on this software, and how to contribute to it, please see:
 *
 * http://thomas.winzig.com/beos/ (or http://www.winzig.com/beos/ if the 1st is down)
 *
 * Copyright (c) 1999 Thomas S. Winzig. All rights reserved.
 *
 * -----------------------------------------------------------------------
 */

#include <Application.h>
#include <Roster.h>
#include <Message.h>
#include <List.h>
#include <Messenger.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "usb/USBKit.h"
#include "USBLooper.h"
#include "USBLooperUtil.h"

USBLooperRoster::USBLooperRoster()
{
	usb_devices = new BMessage(USB_DEVICES);
	usb_devices_paths = new BMessage(USB_DEVICES_PATHS);
	apps = new BList(3);	// holds our list of interested applications
}

USBLooperRoster::~USBLooperRoster()
{
	BMessage message(USB_LOOPER_KABOOM);
	apps->DoForEach(&NotifyApp, &message);
	
	delete apps;
	delete usb_devices_paths;
	delete usb_devices;
	
	#ifdef DEBUG
		fprintf(stderr, "USBLooper Exiting...\n");
	#endif
}

status_t USBLooperRoster::DeviceAdded(USBDevice *dev)
{
	const char* location = dev->Location();
	const char* name = dev->ProductString();
	
	// Get the full path to the device in the var
	char* fullpath = new char[strlen(USB_ROOT) + strlen(location) + 1];
	sprintf(fullpath, "%s%s", USB_ROOT, location);
	
	BMessage message(USB_DEVICE_CONNECTED);
	message.AddString("name", name);
	message.AddString("location", fullpath);
	apps->DoForEach(&NotifyApp, &message);
	
	usb_devices->AddString(name, fullpath);
	usb_devices_paths->AddString(fullpath, name);
	
	#ifdef DEBUG
		if (dev->IsHub())
		{
			fprintf(stderr, "Added hub at %s\n", fullpath);			
		}
		else
		{
			fprintf(stderr, "Added device '%s' at %s\n", 
				usb_devices_paths->FindString(fullpath), fullpath);
		}
	#endif
	
	delete fullpath;
	return B_OK;
}

void USBLooperRoster::DeviceRemoved(USBDevice *dev)
{
	const char* location = dev->Location();
	
	// Get the full path to the device in the var
	char* fullpath = new char[strlen(USB_ROOT) + strlen(location) + 1];
	sprintf(fullpath, "%s%s", USB_ROOT, location);

	const char* name = usb_devices_paths->FindString(fullpath);

	#ifdef DEBUG
		if (dev->IsHub())
		{
			fprintf(stderr, "Removing hub\n");		
		}
		else
		{
			fprintf(stderr, "Removing device '%s'\n", name);
		}
	#endif
	
	BMessage message(USB_DEVICE_DISCONNECTED);
	message.AddString("name", name);
	message.AddString("location", fullpath);
	apps->DoForEach(&NotifyApp, &message);
	
	// TODO: Get index here for the correct device to be removed.
	//usb_devices->RemoveData(name, index); // then erase next line:
	usb_devices->RemoveName(name);
	usb_devices_paths->RemoveName(fullpath);
	
	delete fullpath;
}

void USBLooperRoster::MessageReceived(BMessage *message)
{
	BMessenger remote_app = message->ReturnAddress();
	switch (message->what)
	{
		// This message indicates an application that wants to
		// receive messages when USB devices connect/disconnect.
		case USB_LOOPER_HELLO:
		{
			if (!apps->HasItem(&remote_app))
			{
				if (message->FindBool("high"))
				{
					// An uppety app! High priority. First in line...
					apps->AddItem(&remote_app, 0);
				}
				else
				{
					apps->AddItem(&remote_app);
				}
			}
			break;
		}
		// This message indicates an application that no longer
		// wants messages from us. A restraining order...
		case USB_LOOPER_GOODBYE:
		{
			apps->RemoveItem(&remote_app);
			break;
		}
		// This message is a request for us to send our current
		// list of USB devices that are connected. 
		case USB_DEVICES_ALREADY_CONNECTED:
		{
			message->SendReply(usb_devices);
			break;
		}
	}	
}

USBLooper::USBLooper() : BApplication(USB_LOOPER_SIG)
{
	roster = new USBLooperRoster();
	roster->Start();
}

void USBLooper::MessageReceived(BMessage *message)
{
	// Dispatch directly to the roster
	roster->MessageReceived(message);
}


int main(void)
{
	USBLooper usblooper;
	usblooper.Run();
	return 0;
}
